@@ -17,7 +17,7 @@
 #' new_data_frame(list(x = 1:10, y = 10:1))
 new_data_frame <- function(x = list(), n = NULL, ..., class = character()) {
   stopifnot(is.list(x))
-  n <- n %||% df_length(x)
+  n <- n %||% df_size(x)
   stopifnot(is.integer(n), length(n) == 1L)
 
   # names() should always be a character vector, but we can't enforce that
@@ -44,18 +44,7 @@ data_frame <- function(...) {
 }
 
 #' @export
-vec_restore.data.frame <- function(x, to) {
-  # Copy attribute, preserving existing names & recreating rownames
-  attr_to <- attributes(to)
-  attr_to[["names"]] <- names(x)
-  attr_to[["row.names"]] <- .set_row_names(df_length(x))
-  attributes(x) <- attr_to
-
-  x
-}
-
-#' @export
-vec_ptype_full.data.frame <- function(x) {
+vec_ptype_full.data.frame <- function(x, ...) {
   if (length(x) == 0) {
     return(paste0(class(x)[[1]], "<>"))
   } else if (length(x) == 1) {
@@ -77,23 +66,39 @@ vec_ptype_full.data.frame <- function(x) {
 }
 
 #' @export
-vec_ptype_abbr.data.frame <- function(x) {
+vec_ptype_abbr.data.frame <- function(x, ...) {
   paste0("df", vec_ptype_shape(x))
 }
 
+#' @export
+vec_proxy.data.frame <- function(x, ...) {
+  x
+}
+#' @export
+vec_proxy_compare.data.frame <- function(x, ..., relax = FALSE) {
+  x[] <- lapply(x[], vec_proxy_compare_default, relax = TRUE)
+  x
+}
+
+
 # Coercion ----------------------------------------------------------------
 
 #' @rdname new_data_frame
-#' @export vec_type2.data.frame
-#' @method vec_type2 data.frame
+#' @export vec_ptype2.data.frame
+#' @method vec_ptype2 data.frame
 #' @export
-vec_type2.data.frame <- function(x, y) UseMethod("vec_type2.data.frame", y)
-#' @method vec_type2.data.frame data.frame
+vec_ptype2.data.frame <- function(x, y, ...) UseMethod("vec_ptype2.data.frame", y)
+#' @method vec_ptype2.data.frame data.frame
 #' @export
-vec_type2.data.frame.data.frame <- function(x, y) df_col_type2(x, y)
-#' @method vec_type2.data.frame default
+vec_ptype2.data.frame.data.frame <- function(x, y, ..., x_arg = "x", y_arg = "y") {
+  .Call(vctrs_type2_df_df, x, y, x_arg, y_arg)
+}
+#' @method vec_ptype2.data.frame default
 #' @export
-vec_type2.data.frame.default    <- function(x, y) stop_incompatible_type(x, y)
+vec_ptype2.data.frame.default <- function(x, y, ..., x_arg = "x", y_arg = "y") {
+  vec_default_ptype2(x, y, x_arg = x_arg, y_arg = y_arg)
+}
+
 
 # Cast --------------------------------------------------------------------
 
@@ -101,68 +106,56 @@ vec_type2.data.frame.default    <- function(x, y) stop_incompatible_type(x, y)
 #' @export vec_cast.data.frame
 #' @method vec_cast data.frame
 #' @export
-vec_cast.data.frame <- function(x, to) {
+vec_cast.data.frame <- function(x, to, ...) {
   UseMethod("vec_cast.data.frame")
 }
 #' @export
 #' @method vec_cast.data.frame data.frame
-vec_cast.data.frame.data.frame <- function(x, to) df_col_cast(x, to)
+vec_cast.data.frame.data.frame <- function(x, to, ..., x_arg = "x", to_arg = "to") {
+  .Call(vctrs_df_as_dataframe, x, to, x_arg, to_arg)
+}
 #' @export
 #' @method vec_cast.data.frame default
-vec_cast.data.frame.default    <- function(x, to) stop_incompatible_cast(x, to)
+vec_cast.data.frame.default <- function(x, to, ..., x_arg = "x", to_arg = "to") {
+  vec_default_cast(x, to, x_arg = x_arg, to_arg = to_arg)
+}
+
+#' @export
+vec_restore.data.frame <- function(x, to, ..., n = NULL) {
+  .Call(vctrs_df_restore, x, to, n)
+}
+
+
+# AsIS --------------------------------------------------------------------
+
+# Arises with base df ctor: `data.frame(x = I(list(1, 2:3)))`
+
+#' @export
+vec_proxy.AsIs <- function(x, ...) {
+  class(x) <- setdiff(class(x), "AsIs")
+  vec_proxy(x)
+}
+#' @export
+vec_restore.AsIs <- function(x, to, ...) {
+  I(x)
+}
 
 # Helpers -----------------------------------------------------------------
 
-df_length <- function(x) {
-  if (length(x) > 0) {
-    length(x[[1]])
-  } else {
-    0L
-  }
+df_size <- function(x) {
+  .Call(vctrs_df_size, x)
 }
 
-df_col_type2 <- function(x, y) {
-  # Avoid expensive [.data.frame
-  x_raw <- vec_data(vec_slice(x, 0L))
-  y_raw <- vec_data(vec_slice(y, 0L))
-
-  # Find types
-  names <- set_partition(names(x), names(y))
-  if (length(names$both) > 0) {
-    common_types <- map2(x_raw[names$both], y_raw[names$both], vec_type2)
-  } else {
-    common_types <- list()
-  }
-  only_x_types <- x_raw[names$only_x]
-  only_y_types <- y_raw[names$only_y]
-
-  # Combine and restore order and type
-  out <- c(common_types, only_x_types, only_y_types)
-  out <- out[c(names(x), names$only_y)]
-  vec_restore(out, x)
-}
-
-df_col_cast <- function(x, to) {
-  # Avoid expensive [.data.frame method
-  out <- vec_data(x)
-
-  # Coerce common columns
-  common <- intersect(names(x), names(to))
-  out[common] <- map2(out[common], to[common], vec_cast)
-
-  # Add new columns
-  from_type <- setdiff(names(to), names(x))
-  out[from_type] <- map(to[from_type], vec_na, n = vec_size(x))
-
-  # Drop extra columns
-  out <- out[names(to)]
+df_lossy_cast <- function(out, x, to) {
   extra <- setdiff(names(x), names(to))
-  if (length(extra) > 0 ) {
-    warn_lossy_cast(
-      x, to,
-      details = inline_list("Dropped variables: ", extra, quote = "`")
-    )
-  }
 
-  vec_restore(out, to)
+  maybe_lossy_cast(
+    result = out,
+    x = x,
+    to = to,
+    lossy = length(extra) > 0,
+    locations = int(),
+    details = inline_list("Dropped variables: ", extra, quote = "`"),
+    .subclass = "vctrs_error_cast_lossy_dropped"
+  )
 }
