@@ -9,10 +9,10 @@ test_that("data frames print nicely", {
     file = test_path("test-type-data-frame.txt"),
     {
       cat("mtcars:\n")
-      vec_ptype(mtcars)
+      vec_ptype_show(mtcars)
       cat("\n")
       cat("iris:\n")
-      vec_ptype(iris)
+      vec_ptype_show(iris)
     }
   )
 })
@@ -26,7 +26,7 @@ test_that("embedded data frames print nicely", {
   expect_known_output(
     file = test_path("test-type-data-frame-embedded.txt"),
     {
-      vec_ptype(df)
+      vec_ptype_show(df)
     }
   )
 })
@@ -35,15 +35,15 @@ test_that("embedded data frames print nicely", {
 
 test_that("data frame only combines with other data frames or NULL", {
   dt <- data.frame(x = 1)
-  expect_equal(vec_type_common(dt, NULL), vec_type(dt))
-  expect_error(vec_type_common(dt, 1:10), class = "error_incompatible_type")
+  expect_equal(vec_ptype_common(dt, NULL), vec_ptype(dt))
+  expect_error(vec_ptype_common(dt, 1:10), class = "vctrs_error_incompatible_type")
 })
 
 test_that("data frame takes max of individual variables", {
   dt1 <- data.frame(x = FALSE, y = 1L)
   dt2 <- data.frame(x = 1.5, y = 1.5)
 
-  expect_equal(vec_type_common(dt1, dt2), vec_type_common(dt2))
+  expect_equal(vec_ptype_common(dt1, dt2), vec_ptype_common(dt2))
 })
 
 test_that("data frame combines variables", {
@@ -52,11 +52,18 @@ test_that("data frame combines variables", {
 
   dt3 <- max(dt1, dt2)
   expect_equal(
-    vec_type_common(dt1, dt2),
-    vec_type_common(data.frame(x = double(), y = double()))
+    vec_ptype_common(dt1, dt2),
+    vec_ptype_common(data.frame(x = double(), y = double()))
   )
 })
 
+test_that("empty data frame still has names", {
+  df <- data.frame()
+  out <- vec_ptype_common(df, df)
+
+  expect_equal(names(out), character())
+})
+
 # casting -----------------------------------------------------------------
 
 test_that("safe casts work as expected", {
@@ -72,19 +79,30 @@ test_that("warn about lossy coercions", {
   df1 <- data.frame(x = 1, y = 1)
   df2 <- data.frame(x = c("a", 1), stringsAsFactors = FALSE)
 
-  expect_condition(vec_cast(df1, df1[1]), class = "warning_lossy_cast")
-  expect_condition(vec_cast(df2, df1), class = "warning_lossy_cast")
+  expect_lossy(vec_cast(df1, df1[1]), df1[1], x = df1, to = df1[1])
+  expect_lossy(vec_cast(df2, df1), data.frame(x = dbl(NA, 1), y = dbl(NA, NA)), x = chr(), to = dbl())
+
+  out <-
+    allow_lossy_cast(
+      allow_lossy_cast(
+        vec_cast(df2, df1),
+        chr(), dbl()
+      ),
+      df2, df1
+    )
+
+  expect_identical(out, data.frame(x = dbl(NA, 1), y = dbl(NA, NA)))
 })
 
 test_that("invalid cast generates error", {
-  expect_error(vec_cast(1L, data.frame()), class = "error_incompatible_cast")
+  expect_error(vec_cast(1L, data.frame()), class = "vctrs_error_incompatible_cast")
 })
 
 test_that("column order matches type", {
   df1 <- data.frame(x = 1, y = "a")
   df2 <- data.frame(x = TRUE, z = 3)
 
-  df3 <- vec_cast(df2, vec_type_common(df1, df2))
+  df3 <- vec_cast(df2, vec_ptype_common(df1, df2))
   expect_named(df3, c("x", "y", "z"))
 })
 
@@ -97,9 +115,44 @@ test_that("casts preserve outer class", {
 })
 
 test_that("restore generates correct row/col names", {
-  df1 <- data.frame(x = 1:4, y = 1:4, z = 1:4)
-  df2 <- vec_restore(lapply(df1[1:3], `[`, 1:2), df1)
+  df1 <- data.frame(x = NA, y = 1:4, z = 1:4)
+  df1$x <- data.frame(a = 1:4, b = 1:4)
+
+  df2 <- vec_restore(lapply(df1[1:3], vec_slice, 1:2), df1)
 
   expect_named(df2, c("x", "y", "z"))
   expect_equal(.row_names_info(df2), -2)
 })
+
+test_that("cast to empty data frame preserves number of rows", {
+  out <- vec_cast(new_data_frame(n = 10L), new_data_frame())
+  expect_equal(nrow(out), 10L)
+})
+
+test_that("can cast unspecified to data frame", {
+  df <- data.frame(x = 1, y = 2L)
+  expect_identical(vec_cast(unspecified(3), df), vec_init(df, 3))
+})
+
+test_that("can restore lists with empty names", {
+  expect_identical(vec_restore(list(), data.frame()), data.frame())
+})
+
+test_that("can restore subclasses of data frames", {
+  expect_identical(vec_restore(list(), subclass(data.frame())), subclass(data.frame()))
+  scoped_global_bindings(
+    vec_restore.vctrs_foobar = function(x, to, ..., i) "dispatched"
+  )
+  expect_identical(vec_restore(list(), subclass(data.frame())), "dispatched")
+})
+
+test_that("df_as_dataframe() checks for names", {
+  x <- data_frame(1)
+  y <- data_frame(2)
+  expect_error(vec_cast_common(x, y), "must have names")
+})
+
+test_that("can slice AsIs class", {
+  df <- data.frame(x = I(1:3), y = I(list(4, 5, 6)))
+  expect_identical(vec_slice(df, 2:3), unrownames(df[2:3, ]))
+})
